/****************************************************************************/
/* SELFISHDETOUR:selfishdetour-info */
/*     This file is part of Selfish Detour: OS detour(noise) benchmark
 */
/* SELFISHDETOUR:selfishdetour-info */
/* */
/* SELFISHDETOUR:selfishdetour-fillin */
/*     $Id: check-setitimer-overhead.c,v 1.13 2007/01/03 21:16:49 kazutomo Exp $
 *     SelfishDetour_Version: 1.0.7
 *     SelfishDetour_Heredity: FOSS_ORIG
 *     SelfishDetour_License: GPL
 */
/* SELFISHDETOUR:selfishdetour-fillin */
/* */
/* SELFISHDETOUR:selfishdetour-gpl */
/*      Copyright: Argonne National Laboratory, Department of Energy,
 *                 and University of Chicago.  2004, 2005, 2006, 2007
 *      SelfishDetour License: GPL
 * 
 *      This software is free.  See the file license-files/license.GPL
 *      for complete details on your rights to copy, modify, and use this
 *      software.
 */
/* SELFISHDETOUR:selfishdetour-gpl */
/****************************************************************************/
/* -- */
/** 
 * This program measures BG/L CNK interval timer overhead.
 * To measure, this program use 3 different types of test
 *
 * 1. measure the elapsed time for a loop iterates 10 MFLOPS without
 *    interval timer. In other words, measure the time for 10 MLOPS
 *    wihtout any interrupt.
 *
 * 2. measure the elapsed time for a 10MFLOPS loop with 
 *    interval timer. signal is issued every 1 ms. The signal
 *    handler does only increment handler_cnt.
 *
 * With above 2 tests, we could find an average overhead at each signal.
 *
 * 3. similar as test 2. but signal handler is different. In this test,
 *    signal handler calls setitimer() to change interval. Accually,
 *    interval is set altenatively to interval + 1 or interval - 1.
 *    It's equivalent to interval. We just want to avoid non-update 
 *    situation.
 *
 * At this point, we could find an average overhead of new timer value
 * update with setitimer( ).
 * 
 * Kazutomo Yoshii <kazutomo@mcs.anl.gov>
 *
 */

/* ---------

* DATE: 06/Oct/2006 / DRIVER:V1R3M0

[without timer interrupt]
elapsed_ticks=60000097

[with timer interrupt]
elapsed_ticks=60608231
handler_cnt=87

[with timer interrupt + timer renew everytime ]
elapsed_ticks=60847950
handler_cnt=86

signal_overhead=6990 ticks / 10.000000 usec
setitimer_overhead=2868 ticks / 4.103004 usec



* DATE: 06/MAR/2006 / DRIVER:V1R2M1

[without timer interrupt]
elapsed_ticks=220000134

[with timer interrupt]
elapsed_ticks=225460576
handler_cnt=322

[with timer interrupt + timer renew everytime ]
elapsed_ticks=226316351
handler_cnt=317

signal_overhead=16957 ticks / 24.258941 usec
setitimer_overhead=2967 ticks / 4.244635 usec


* DATE:25/OCT/2005 / DRIVER:??

[without timer interrupt]
elapsed_ticks=220000116

[with timer interrupt]
elapsed_ticks=223467831
handler_cnt=319

[with timer interrupt + timer renew everytime ]
elapsed_ticks=224346877
handler_cnt=315

signal_overhead=10870 ticks / 15.550787 usec
setitimer_overhead=2929 ticks / 4.190272 usec

  ---------- */

#include "selfish_utils.h"  /* includes most necessary header */


#define   X_FLOPS          (1e+07)     // 10 MFLOPS
#define   FIXED_INTERVAL   (1000)      // 1ms


#if defined(__blrts__)  
#define   TICKS_PER_USEC   (699)
#else
static int TICKS_PER_USEC;
#endif

volatile uint64_t handler_cnt=0;

#ifndef BLRTS_DECREMENTER_HIJACK
static void handler(int sig)
{
  handler_cnt ++;
}

static void handler_resettimer(int sig)
{
  struct itimerval itv;
  int rc;
  static int flip=-1;

  handler_cnt ++;

  if( flip < 0 ) flip = +1;
  else           flip = -1;
  itv.it_value.tv_sec   = 0;
  itv.it_value.tv_usec  = FIXED_INTERVAL+flip;      // 1ms +- 1us 
  itv.it_interval       = itv.it_value;

  /* 
     the timer decrements from it_value to zero,
     and signal is issued
     reset to it_interval   (right after signal has been issued ? )
  */
  rc = setitimer(ITIMER_REAL, &itv, 0);
  assert( rc == 0 );
}
#else
/* for BLRTS DECREMENTER HIJICK */
#include <cnkhack.h>

static void decrementer_handler(void)
{
  handler_cnt ++;
}
#endif



int main(int agrc, char* argv[])
{
  struct itimerval itv;
  int rc;
  uint64_t start,stop;
  double x;
  uint64_t elapsed0,elapsed1,elapsed2;
  uint64_t handler_cnt1,handler_cnt2;
  uint64_t signal_overhead;
  uint64_t setitimer_overhead;

#ifndef __blrts__
  TICKS_PER_USEC = (cpufreq()/1000)/1000;
#endif

  /* -----------------------
     Without timer interrupt
     ----------------------- */

  /* warming up */
  handler_cnt = 0;  x=0.0;
  start = rdtsc();
  while( x < X_FLOPS ) {  x = x + 1.0;  }  /* warm up */
  stop = rdtsc();

  handler_cnt = 0;  x=0.0;
  start = rdtsc();
  while( x < X_FLOPS ) {  x = x + 1.0;  }
  stop = rdtsc();

  elapsed0 = stop-start;

  /* -----------------------
     With timer interrupt
     ----------------------- */
  itv.it_value.tv_sec   = 0;
  itv.it_value.tv_usec  = FIXED_INTERVAL;
  itv.it_interval       = itv.it_value;

  /* 
     the timer decrements from it_value to zero,
     and signal is issued
     reset to it_interval   (right after signal has been issued ? )
  */
#ifndef BLRTS_DECREMENTER_HIJACK
  signal( SIGALRM, handler);
#else
  hijack_decrementer_handler(decrementer_handler);
#endif

  rc = setitimer(ITIMER_REAL, &itv, 0);
  assert( rc == 0 );


  /* 
    then, do the fixed amount of work here
  */
  handler_cnt = 0;  x=0.0;
  start = rdtsc();
  while( x < X_FLOPS ) {  x = x + 1.0;  }  /* warm up */
  stop = rdtsc();

  handler_cnt = 0;  x=0.0;
  start = rdtsc();
  while( x < X_FLOPS ) {  x = x + 1.0;  }
  stop = rdtsc();

  elapsed1 = stop-start;
  handler_cnt1 = handler_cnt;


#ifndef BLRTS_DECREMENTER_HIJACK
  /* -----------------------
     With timer interrupt + timer renew everytime
     ----------------------- */
  signal( SIGALRM, handler_resettimer );
  handler_cnt = 0;  x=0.0;
  start = rdtsc();
  while( x < X_FLOPS ) {  x = x + 1.0;  }  /* warm up */
  stop = rdtsc();

  handler_cnt = 0;  x=0.0;
  start = rdtsc();
  while( x < X_FLOPS ) {  x = x + 1.0;  }
  stop = rdtsc();

  elapsed2 = stop-start;
  handler_cnt2 = handler_cnt;
#endif


  /* =====================
     REPORTING RESULT
     =====================  */
  printf("[without timer interrupt]\n");
  printf("elapsed_ticks=%llu\n", elapsed0);
  printf("\n");


  printf("[with timer interrupt]\n");
  printf("elapsed_ticks=%llu\n", elapsed1);
  printf("handler_cnt=%llu\n", handler_cnt1);
  printf("\n");

#ifndef BLRTS_DECREMENTER_HIJACK
  printf("[with timer interrupt + timer renew everytime ]\n");
  printf("elapsed_ticks=%llu\n", elapsed2);
  printf("handler_cnt=%llu\n", handler_cnt2);
  printf("\n");
#endif


  signal_overhead = (elapsed1-elapsed0)/handler_cnt1;
  setitimer_overhead = (elapsed2-elapsed0)/handler_cnt2 - signal_overhead;    

  printf("signal_overhead=%llu ticks / %f usec\n", 
	 signal_overhead, 
	 (double)signal_overhead/(double)TICKS_PER_USEC  );
  
#ifndef BLRTS_DECREMENTER_HIJACK
  printf("setitimer_overhead=%llu ticks / %f usec\n",
	 setitimer_overhead,
	 (double)setitimer_overhead/(double)TICKS_PER_USEC );
#endif

  // for(;;) ;

  return 0;
}

/*  $RCSfile: check-setitimer-overhead.c,v $   $Author: kazutomo $
 *  $Revision: 1.13 $  $Date: 2007/01/03 21:16:49 $
 */

