/****************************************************************************/
/* SELFISHDETOUR:selfishdetour-info */
/*     This file is part of Selfish Detour: OS detour(noise) benchmark
 */
/* SELFISHDETOUR:selfishdetour-info */
/* */
/* SELFISHDETOUR:selfishdetour-fillin */
/*     $Id: record_selfish_detour.c,v 1.45 2007/01/03 21:16:49 kazutomo Exp $
 *     SelfishDetour_Version: 1.0.7
 *     SelfishDetour_Heredity: FOSS_ORIG
 *     SelfishDetour_License: GPL
 */
/* SELFISHDETOUR:selfishdetour-fillin */
/* */
/* SELFISHDETOUR:selfishdetour-gpl */
/*      Copyright: Argonne National Laboratory, Department of Energy,
 *                 and University of Chicago.  2004, 2005, 2006, 2007
 *      SelfishDetour License: GPL
 * 
 *      This software is free.  See the file license-files/license.GPL
 *      for complete details on your rights to copy, modify, and use this
 *      software.
 */
/* SELFISHDETOUR:selfishdetour-gpl */
/****************************************************************************/
/*
 * record_selfish_detour [options]
 * 
 * check usage() for option
 */

#include "selfish_detour.h"

#ifdef MPI_MODE
#include "mpi.h"
#endif


#define RECORD_SELFISH_DETOUR_VERSION  DETOUR_VERSION


#include <stdarg.h>
#ifdef MPI_MODE
static int local_printf(const char* fmt, ...)
{
  int rank;
  int ret=0;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  if( rank==0 ) {
    va_list ap;
    va_start(ap,fmt);
    ret = vprintf(fmt, ap);
    va_end(ap);
  }
  return ret;
}
static int local_puts(const char* buf)
{
  int rank;
  int ret=0;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  if( rank==0 ) {
    ret = puts(buf);
  }
  return ret;
}
#else

static int local_printf(const char* fmt, ...)
{
  int ret;
  va_list ap;
  va_start(ap,fmt);
  ret = vprintf(fmt, ap);
  va_end(ap);
  return ret;
}
static int local_puts(const char* buf)
{
  int ret=0;
  ret = puts(buf);
  return ret;
}
#endif



/**  
 * detour_segs[i*2+0] : start tick of i-th detour (left edge)
 * detour_segs[i*2+1] : end   tick of i-th detour (right edge)  
 */
static uint64_t* detour_segs;

/**
 * the following variables are configurable by the command line option 
 */

static int N_SEGS = 1000;
static char* output_fn_base = "detour";
static char* output_fn_ext  = ".dat";
static uint64_t precision_ns = 1000;     /* in nano second */
static uint64_t precision_ticks = -1;    /* = cpufreq()*(precision_ns * 10^-9) */

static int    fake_detour_interval_us   = -1;
static double fake_detour_prob_or_count = 0.0;
static int    fake_detour_duration_us   = -1;
static int    fake_detour_sync_flag     = -1;


static int    use_second_for_output     = 0;


/** 
 * the following variables are intialized and updated by acquisition()
 */
static uint64_t min_loop_ticks=-1;
static uint64_t start_tick=-1;
static uint64_t stop_tick=-1;
static uint64_t timeout_tick=UINT64_MAX;

static int   idxcnt=-1;   /**
			   * idxcnt/2 is the number of detours recorded 
			   * Usually idxcnt/2 is same as N_SEGS
			   * idxcnt/2 < N_SEGS when timeout'ed
			   */


static void usage() 
{
  char* usage_msg = 
    "\n"  
    "record_selfish_detour [options]\n" 
    "\n" 
    "-s datasize    : maximum recording buffer length  (1000)\n" 
    "-o outputfn    : data output file basename (detour.dat) \n" 
    "-p precision   : threashold value for detour in nano second (1000 nano sec)\n" 
    "-t timeoutsec  : timeout value in second (120 sec)\n" 
    "-u             : use second instead of cpu tick value for output file (off)\n" 
    "-M timeoutsec  : measure the max latency without storing data into array (off)\n"
    "-h             : help message\n" 
    "-V             : version information\n" 
    "\n"
    "* () contains default value or setting\n"
    "\n" 
#if defined(__blrts__)
    "BlueGene/L BLRTS only features\n" 
    "\n" 
    "-F i,p,d,s     : generate fake detour during recording with noise injection\n"
    "                 the purpose of this is to check noise injection library\n"
    "           i -> interval in micro second between fake detours\n" 
    "           p -> probability (0.0 to 1.0) or counter(p>1)\n" 
    "           d -> duration in micro second for each fake detours\n" 
    "           s -> if '1' is specified , sync noise (no seed)\n" 
    "  ex.\n" 
    "  -F 1000,1.0,0,0  means min. noise every 1ms\n" 
    "  -F 1000,100,80,0 means min. noise every 1ms + 80usec noise every 100 times\n" 
    "  -F 1000,0.1,80,0 means noise occurs every 1ms, \n" 
    "                         10% of 80usec and 90% of min noise \n"
#endif
    "\n";

  local_printf(usage_msg);
}

static void acquisition(void)
{
  uint64_t current, prev;

  assert( precision_ticks > 0 );

  /**
   * acquisition loop start
   */

  idxcnt = 0;
  min_loop_ticks = cpufreq();

  /* 
   * main acquisition loop 
   */

#ifdef MPI_MODE
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  start_tick = current = rdtsc();

  while( (idxcnt < (N_SEGS*2))  ) {

    prev = current;
    current = rdtsc();   /* get the current cpu tick counter value */

    if( current - prev > precision_ticks ) {
      detour_segs[idxcnt++] = prev;
      detour_segs[idxcnt++] = current;
    } 

    if( (current - prev) < min_loop_ticks ) {
      /* update minimum ticks */
      min_loop_ticks = current - prev; 
    }

    if( !( (current-start_tick)<timeout_tick )  )  break;
  }
  stop_tick = rdtsc();

  /*
   * main acquisition loop end
   */

  if( (current-start_tick)>=timeout_tick ) {
    int i;
    local_printf("timeout'ed. idxcnt=%d, min_loop_ticks=%llu\n", 
	   idxcnt, min_loop_ticks);
    if( idxcnt < 10 ) {
      for( i=0; i<idxcnt/2; i++ ) 
	local_printf("%llu %llu\n", 
	       detour_segs[i*2+0], detour_segs[i*2+1] ); 
    }
  }
}


static void write_to_file(void)
{
  FILE* fp;
  uint64_t td,tdmax = 0, tdmin = cpufreq();
  int i;
  int n_segs;
  char output_fn[128];

  /* caclculate the max latency and the min latency */


  n_segs = idxcnt / 2;

  local_printf("N_SEGS=%d  idxcnt/2=%d\n", N_SEGS, n_segs);

  for( i=0; i<n_segs; i++ ) {
    td = detour_segs[i*2+1] - detour_segs[i*2+0] ;
    if( td > tdmax ) tdmax = td;
    if( td < tdmin ) tdmin = td;
  }

  if( tdmin < min_loop_ticks ) { 
     local_printf("Warning: tdmin(=%llu) < min_loop_ticks(=%llu)\n",
 	tdmin, min_loop_ticks  );
  }

#ifdef MPI_MODE
 {
   int rank;
   MPI_Comm_rank(MPI_COMM_WORLD, &rank);
   snprintf(output_fn, sizeof(output_fn), "%s%04d%s", 
	    output_fn_base, rank, output_fn_ext);
 }

#else
  snprintf(output_fn, sizeof(output_fn), "%s%s",
	   output_fn_base, output_fn_ext );
#endif  

  fp = fopen(output_fn, "w");
  assert( fp != (FILE*)NULL );

  fprintf(fp, "# \n");
  fprintf(fp, "# [Parameters]\n");
  fprintf(fp, "# \n");
  fprintf(fp, "# CPU_FREQ_HZ=%llu\n", cpufreq());
  fprintf(fp, "# START_TICK=%llu\n", start_tick);
  fprintf(fp, "# LAST_TICK=%llu\n", stop_tick);
  fprintf(fp, "# MIN_TICKS=%llu\n", min_loop_ticks);
  fprintf(fp, "# PRECISION=%llu\n", precision_ticks);
  fprintf(fp, "# MAX_LATENCY=%llu\n", tdmax);
  fprintf(fp, "# MIN_LATENCY=%llu\n", tdmin);
  fprintf(fp, "# \n");
  fprintf(fp, "# N_SEGS=%d\n", n_segs);
  fprintf(fp, "# \n");
  fprintf(fp, "# TOOL_VERSION=%s\n", RECORD_SELFISH_DETOUR_VERSION);
  fprintf(fp, "# \n");
#ifdef MPI_MODE
  fprintf(fp, "# MPI_MODE # data acquisition is started with synchronization manner.\n");
  fprintf(fp, "# \n");
#endif
  fprintf(fp, "# DETOUR\n");
  if( use_second_for_output ) {
    fprintf(fp, "# START(sec)  DURATION(sec)\n");
  } else {
    fprintf(fp, "# START(tick)  DURATION(tick)\n");
  }


  for( i=0; i<n_segs; i++ ) {
    td = detour_segs[i*2+1] - detour_segs[i*2+0] ;

    if( td > tdmax ) tdmax = td;

    if( use_second_for_output ) {
      fprintf(fp, "%14.9f %12.9f\n",
	      ((double)detour_segs[i*2+0] - (double)start_tick)/(double)cpufreq() ,
	      (double)td/(double)cpufreq() 
	      );
    } else {
      fprintf(fp, "%llu %llu\n",
	      detour_segs[i*2+0] - start_tick,
	      td  );
    }
  }

  fclose(fp);
}


static void timeouted(int sig)
{
  local_printf("warning: timeout'ed but writing recording data to file\n");
  write_to_file();
  local_printf("done\n");
  
  exit(0);
}


/* variables for report_latency() */
static uint64_t maximum_latency=0;
static double _x_counter=0;

static void print_latency(int sig)
{
  local_puts("* results");
  local_printf("maximum latency = %llu\n",maximum_latency);
  local_printf("_x_counter      = %f\n", _x_counter);
  local_printf("cpu frequency   = %llu\n", cpufreq());
  exit(0);
}

static void report_latency(int timeout)
{
  uint64_t current, prev;

  cpufreq();

  local_puts("");
  local_puts("[Maximum latency measurement]");
  local_puts("");
  local_printf("timeout = %d sec\n", timeout);
  local_puts("");
  local_puts("* benchmark loop");
  local_puts("maximum_latency=0;");
  local_puts("_x_counter=0.0");
  local_puts("current = rdtsc();");
  local_puts("for(;;) {");
  local_puts("  prev = current;");
  local_puts("  current = rdtsc();");
  local_puts("  if( current - prev > maximum_latency ) {");
  local_puts("     maximum_latency = current - prev;");
  local_puts("  }");
  local_puts("  _x_count += 1.0");
  local_puts("}");
  local_puts("");

  /* warm up */
  maximum_latency=0;
  _x_counter=0.0;
  current = rdtsc();
  for(;_x_counter<10000;) {
    prev = current;
    current = rdtsc();
    if( current - prev > maximum_latency ) {
      maximum_latency = current - prev;
    }
    _x_counter += 1.0;
  }


  signal(SIGALRM, print_latency );
  alarm(timeout);   

  /*  this checks only latency and perfomance */
  maximum_latency=0;
  _x_counter=0.0;
  current = rdtsc();
  for(;;) {
    prev = current;
    current = rdtsc();
    if( current - prev > maximum_latency ) {
      maximum_latency = current - prev;
    }
    _x_counter += 1.0;
  }
}

int main(int argc, char* argv[])
{
  int i;
  int c;
  int timeout=120;

#ifdef MPI_MODE
  MPI_Init(&argc, &argv);
#endif

  for(;;) {
    c = getopt( argc, argv, "M:s:p:o:t:F:Vhu");

    if (c == EOF) break;
    switch (c) {
    case 'M':
      report_latency(atoi(optarg));
      exit(0);
      break;
    case 'F':
      {
	char tmp[80];
	char* p;
	int    tmp_interval=-1;
	double tmp_prob_or_count=0.0;
	int    tmp_duration=-1;
	double tmp_sync_flag = 0.0;

	strncpy( tmp, optarg, sizeof(tmp) );
	p = strtok(tmp, ",");
        if( p ) {
          tmp_interval=atoi(p);
          p = strtok(NULL, ",");
	  if( p ) {
	    tmp_prob_or_count = atof(p);
	    p = strtok(NULL, ",");
	    if( p ) {
	      tmp_duration=atoi(p);
	      p = strtok(NULL, ",");
	      if( p ) {
		if( atoi(p) > 0 ) tmp_sync_flag = 1;
	      }
	    }
          }
        }

	if( tmp_interval>0 ) {
	  fake_detour_interval_us   = tmp_interval;
	  fake_detour_prob_or_count = tmp_prob_or_count;
	  fake_detour_duration_us   = tmp_duration;
	  fake_detour_sync_flag     = tmp_sync_flag;
	}
      }

      break;
    case 's':
      {
	int n_segs = atoi( optarg );
	if( n_segs > 0 ) {
	  N_SEGS = n_segs; 
	}
      }
      break;
    case 't':
      timeout = atoi( optarg );
      break;
    case 'p':
      {
	uint64_t _p_ns=0;
	char tmpbuf[80];
	int tmpbuflen;

	strncpy(tmpbuf, optarg, sizeof(tmpbuf));
	tmpbuflen = strlen(tmpbuf);

	if( strlen(tmpbuf) >= 2 &&
	    tmpbuf[tmpbuflen-1] == 's') {

	  switch( tmpbuf[tmpbuflen-2] ) {
	  case 'n' : 
	    tmpbuf[tmpbuflen-2] = 0;
	    _p_ns = atoi(tmpbuf);
	    break;
	  case 'u' : 
	    tmpbuf[tmpbuflen-2] = 0;
	    _p_ns = atoi(tmpbuf) * 1000;
	    break;
	  case 'm' :
	    tmpbuf[tmpbuflen-2] = 0;
	    _p_ns = atoi(tmpbuf) * 1000 * 1000;
	    break ;
	  default:
	    tmpbuf[tmpbuflen-1] = 0;
	    _p_ns = atoi(tmpbuf) * 1000 * 1000 * 1000;
	  }
	} else {
	  _p_ns = atoi(optarg);
	}
	if( _p_ns > 0 ) {
	  precision_ns = _p_ns;
	}

      }
      break;

    case 'o':
      output_fn_base = optarg;
      break;
    case 'V':
      local_printf("record_selfish_detour - version %s\n",RECORD_SELFISH_DETOUR_VERSION);
      local_printf("compile date : %s %s\n", __DATE__,__TIME__);
      exit(0);
      break;
    case 'u':
      use_second_for_output = 1;
      break;
    case 'h':
      usage();
      exit(0);
      break;
    }
  }


  precision_ticks = cpufreq() * precision_ns / 1000 / 1000 / 1000 ;

  local_printf("Detour recording tool  v%s\n", RECORD_SELFISH_DETOUR_VERSION);

  local_printf("[Information]\n");
  local_printf("detected cpufreq : %llu HZ\n", cpufreq());
  local_printf("precision        : %llu nsec  (%llu ticks)\n",precision_ns, precision_ticks);
  local_printf("recording size   : %d\n",N_SEGS);
  local_printf("output filename  : %s\n",output_fn_base);
  local_printf("timeout          : %d sec (0 means no timeout)\n", timeout);
  

  if( fake_detour_interval_us>0 ) {
    local_printf("fake detour      : interval=%d us / probability(or count)=%f / duration=%d us / sync=%d\n",
		 fake_detour_interval_us,
		 fake_detour_prob_or_count,
		 fake_detour_duration_us ,
		 fake_detour_sync_flag );
  }

  detour_segs = (uint64_t*)malloc(N_SEGS*2*sizeof(uint64_t));
  if( ! detour_segs  ) {
    fprintf(stderr, "Unable to allocate memory\n");
    exit(1);
  }


  /* XXX: fake detour is currently available only under MPI_MODE */
  if( fake_detour_interval_us > 0 ) {
    char tmpbuf[20];

    /* noise injection paramteres are passed by env variables.  this
       may be easiest way when we think about fortran linking */
    snprintf(tmpbuf, sizeof(tmpbuf), "%d", fake_detour_interval_us);
    setenv("SELFISH_I", tmpbuf, 1 );
    if( fake_detour_prob_or_count <= 1.0 ) {
      snprintf(tmpbuf, sizeof(tmpbuf), "%f", fake_detour_prob_or_count );
      setenv("SELFISH_P", tmpbuf, 1 );
    } else {
      snprintf(tmpbuf, sizeof(tmpbuf), "%d", (int)fake_detour_prob_or_count );
      setenv("SELFISH_CNT", tmpbuf, 1 );
    }
    snprintf(tmpbuf, sizeof(tmpbuf), "%d", fake_detour_duration_us);
    setenv("SELFISH_D", tmpbuf, 1 );

    if( fake_detour_sync_flag>0 ) {
      snprintf(tmpbuf, sizeof(tmpbuf), "1");
      setenv("SELFISH_SYNC", tmpbuf, 1 );
    }

#if defined(__blrts__)
    { 
      int rank;
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      selfish_detour_init();
    }
#endif 
  } else {
    if( timeout>0 ) signal(SIGALRM, timeouted);
  }

  for(i=0; i<2; i++) {
    local_printf("warm up #%d start\n",i);

    if( timeout>0 ) timeout_tick=cpufreq()*timeout;
    else            timeout_tick=UINT64_MAX;


    acquisition( );  /* warm up */


    if( timeout>0 ) timeout_tick=cpufreq()*timeout;
    else            timeout_tick=UINT64_MAX;

    local_printf("warm up #%d end\n",i);
  }


  local_printf("acquisition start\n");

  if( timeout>0 ) timeout_tick=cpufreq()*timeout;
  else            timeout_tick=UINT64_MAX;

#if defined(__blrts__)
  reset_detour_stats();
#endif

  acquisition( );  

#if defined(__blrts__)
  /* XXX: fake detour is currently available only under MPI_MODE */
  if( fake_detour_interval_us > 0 ) {
    double global_flop;
    unsigned long long  detour_cnt, elapsed_ticks;

    get_detour_stats( &global_flop,
		      &detour_cnt,
		      &elapsed_ticks);

    local_printf("# FAKE_BUSYLOOP_CNT=%d\n",     (unsigned)global_flop);
    local_printf("# FAKE_DETOUR_SIGNAL_COUNT=%llu\n",  detour_cnt);
    local_printf("# FAKE_ELAPSED_TICKS=%llu\n", elapsed_ticks);
    selfish_detour_finalize();
  }
#endif

  local_printf("acquisition end\n");

  write_to_file( );


  local_printf("Acquisition data has been saved to \"%s\".\n", output_fn_base);
  local_printf("# grep '^#' %s\n", output_fn_base);

  local_printf("done.\n");

#ifdef MPI_MODE
  MPI_Finalize();
#endif


  return 0;
}

/*  $RCSfile: record_selfish_detour.c,v $   $Author: kazutomo $
 *  $Revision: 1.45 $  $Date: 2007/01/03 21:16:49 $
 */
