/****************************************************************************/
/* SELFISHDETOUR:selfishdetour-info */
/*     This file is part of Selfish Detour: OS detour(noise) benchmark
 */
/* SELFISHDETOUR:selfishdetour-info */
/* */
/* SELFISHDETOUR:selfishdetour-fillin */
/*     $Id: selfish_detour_prob.c,v 1.20 2007/01/03 21:16:49 kazutomo Exp $
 *     SelfishDetour_Version: 1.0.7
 *     SelfishDetour_Heredity: FOSS_ORIG
 *     SelfishDetour_License: GPL
 */
/* SELFISHDETOUR:selfishdetour-fillin */
/* */
/* SELFISHDETOUR:selfishdetour-gpl */
/*      Copyright: Argonne National Laboratory, Department of Energy,
 *                 and University of Chicago.  2004, 2005, 2006, 2007
 *      SelfishDetour License: GPL
 * 
 *      This software is free.  See the file license-files/license.GPL
 *      for complete details on your rights to copy, modify, and use this
 *      software.
 */
/* SELFISHDETOUR:selfishdetour-gpl */
/****************************************************************************/
/* ===================================
   SELFISH DETOUR NOISE INJECTION CODE

   generate randomized noise with specified probablity 
   =================================== */
#include <stdlib.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>

#include <stdarg.h>
#include <mpi.h>

#include "selfish_utils.h"
#include "selfish_detour.h"

#ifdef BLRTS_DECREMENTER_HIJACK
#include <cnkhack.h>
#endif



static int local_printf(const char* fmt, ...)
{
  int rank;
  int ret=0;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  if( rank==0 ) {
    va_list ap;
    va_start(ap,fmt);
    ret = vprintf(fmt, ap);
    va_end(ap);
  }
  return ret;
}




__attribute__((section(".selfishdata")))
     int selfish_detour_verbose=1;

//**********************************************************************
// for the super simple pre-release version of this code, we just
// statically define several different setups, and they are chosen by the
// parameter passed into this setup routine
//
// column 0 is the "Period" in microsecs between each interruption
// column 1 is "Probability" of a lengthy interruption at the end of a period
// column 2 is the "Length" of the detour in microsecs

static const double detour_config[12][3] = { 
  {1000, 0.0, 1}, 
  {1000, .01, 100}, 
  {1000, .01, 200}, 
  {1000, .01, 400}, 
  {1000, .01, 800}, 
  {1000, .01, 1600}, 
  {1000, .01, 3200}, 
  {1000, .01, 6400}, 
  {1000, .01, 12800}, 
  {1000, .01, 25600}, 
  {1000, .01, 51200},
  {1000, .01, 102400} 
};

/* these value are set by selfish_detour_init() and used by timer
   signal handler */

/* one application used to corrupt memory in .sbss section and added
   memory fench for just in case.  here is a workaround for that.  of
   course, usually it's not necessary
*/

static unsigned stupidfence[1024*1024];  /* apps .sbss */

__attribute__((section(".selfishdata")))
     unsigned data[1024*1024];

__attribute__((section(".selfishdata")))
     unsigned __g1;
__attribute__((section(".selfishdata")))
     double noise_interval_usec;
__attribute__((section(".selfishdata")))
     unsigned __g2;
__attribute__((section(".selfishdata")))
     double noise_probability;
__attribute__((section(".selfishdata")))
     double noise_duration_usec;

__attribute__((section(".selfishdata")))
     unsigned long long selfish_detour_last_tick;

__attribute__((section(".selfishdata")))
     int detour_sync=0;

__attribute__((section(".selfishdata")))
     int noise_decrementer ;

__attribute__((section(".selfishdata")))
     int noise_counter ;

/* if noise_counter has non-zero value, noise injection routine is
   going to be counter mode.  With this mode, noise_decrementer is
   initialized with noise_counter value.  noise_decrementer is
   decremented at each signal.  noise will be injected when
   noise_decrementer gets 1.  This noise_counter is set via
   environment variable.
*/


double get_detour_noise_interval_usec()
{
  return noise_interval_usec;
}

double get_detour_noise_probability()
{
  return noise_probability;
}

double get_detour_noise_duration_usec()
{
  return noise_duration_usec;
}

int get_detour_noise_counter()
{
  return noise_counter;
}

int get_detour_noise_sync()
{
  return detour_sync;
}


/* put data and code that are invoked from timer interrupt into
   special section so we can separate from application data and code */

__attribute__((section(".selfishdata")))
     double global_flop;

__attribute__((section(".selfishdata")))
     unsigned long long detour_signal_cnt;

__attribute__((section(".selfishdata")))
     unsigned long long detour_elapsed_ticks;


void get_detour_stats( double * bogoflop, 
		       unsigned long long* counter,
		       unsigned long long* elapsed_ticks)
{
  *bogoflop = global_flop;
  *counter = detour_signal_cnt;
  *elapsed_ticks = detour_elapsed_ticks;
}

void reset_detour_stats()
{
  global_flop = 0.0;
  detour_signal_cnt = 0;
  detour_elapsed_ticks = 0;
}


/* this value is pre-calculated before timer start */
__attribute__((section(".selfishdata")))
     unsigned long detour_loop_ticks ;
__attribute__((section(".selfishdata")))
     unsigned long noise_probability_uint32;


			
/*
  signal handler for interval timer, which adds noise to app
  
  this function access the following symbols

  detour_loop_ticks
  noise_probability_uint32

  noise_counter
  noise_decrementer

  detour_signal_cnt
  global_flop
  detour_elapsed_ticks
*/
void selfish_detour(int sig) __attribute__((section(".selfishcode")));


#ifndef BLRTS_DECREMENTER_HIJACK
#define panic_printf     printf
#else
#define panic_printf     jtagmbox_printf_direct
#endif

__attribute__((section(".selfishdata")))
     static unsigned long selfish_too_short_interval_count;



void selfish_detour(int sig) 
{
  unsigned long x;
  unsigned long long b,s;


  s = rdtsc();

  /* do consistency check */
  if( (detour_signal_cnt>0) &&
      (s - selfish_detour_last_tick) < (unsigned long long)((noise_interval_usec/10.0) * TICKS_PER_US) ) {
    selfish_too_short_interval_count++;
  }

  //if( selfish_too_short_interval_count > 10 ) {
  if(0) {
    panic_printf(
		 "It's unlikely! interval(%llu us) is too short  (should be around %f) cnt=%llu last=%llu current=%llu (%08x,%08x) sig=%d\n",
		 (s - selfish_detour_last_tick)/TICKS_PER_US, 
		 noise_interval_usec ,
		 detour_signal_cnt,
		 selfish_detour_last_tick,     s,
		 __g1,__g2, sig );

#ifndef BLRTS_DECREMENTER_HIJACK
    exit(1);
#else
    /* cause DTLB fault intentionally to stop the job*/
    /* this asm code is for debugging */
    __asm__ __volatile__ 
      ( "mr 20, %0 \n\t" \
	"mr 21, %1 \n\t" \
	"mr 22, %2 \n\t" 
	: 
	: "r"(__g1), "r"(__g2), "r"(sig) );

    *((unsigned*)0xeeeee000) = 0;
#endif
  }


  detour_signal_cnt++;

  if( detour_loop_ticks == 0 )    goto out;

  x = (unsigned long) randomMT(); 


  /* 
     either decrementer reached to one or hits to the certain probablity,
     it takes detour(busyloop)
   */
  if(  ((noise_counter>0) && noise_decrementer==1) ||
       noise_probability_uint32 > x) {

    b = s;
  
    /* detour or noise */
    while ( (b-s) < detour_loop_ticks) {
      global_flop ++;
      b = rdtsc();
    }

    /* refill the decrementer */
    if( noise_counter>0 ) {
      noise_decrementer = noise_counter;
    }
  } else {
    if( noise_counter>0 ) {
      noise_decrementer --;
    }
  }
 out:
  /* update the stats counter */
  b = rdtsc();
  detour_elapsed_ticks += (b-s);
  selfish_detour_last_tick=s;
}





//********************************************************************
// Env-variable based input processing
int selfish_detour_env()
{
  char* p;
  int   specified_by_env=0;

  p = getenv("SELFISH_VERBOSE");
  if( p ) {    selfish_detour_verbose = atoi(p);   }

  p = getenv("SELFISH_SYNC");
  if( p ) { 
    detour_sync = 1; 
  }

  p = getenv("SELFISH_I"); //INTERVAL (usec)
  if(p) {
    noise_interval_usec = strtod(p, NULL);
    specified_by_env++;
  }

  p = getenv("SELFISH_CNT");
  if( p ) { 
    noise_decrementer= noise_counter = atoi(p); 
  } else {
    p = getenv("SELFISH_P"); //PROB
    if( p ) {
      noise_probability   = strtod(p, NULL);
    }
  }

  p = getenv("SELFISH_D"); //DETOUR  (usec)
  if( p ) {
    noise_duration_usec = strtod(p, NULL);
  }

  if( specified_by_env) return 1;

  return -1;
}





//********************************************************************
// Simple timer/alarm setup
// to make it to be able to call from fortran code
// this function uses the pass-by-reference, instead of the pass-by-value
//
// XXX: we should change API!!!!  val and rank doesn't take effect!!
//
void __selfish_detour_init();

void selfish_detour_init(int *val, int *rank) 
{
  /* 
     read paramters from the environment variable
  */
  selfish_detour_env();

  __selfish_detour_init();
}


void selfish_detour_init_params(double interval_usec, 
				double duration_usec,
				double probability,
				int counter,
				int sync ) 
{
  if( sync > 0 ) detour_sync = 1; 

  if( interval_usec>= 0.0 ) {
    noise_interval_usec = interval_usec;
  }
  if( counter > 0 ) {
    noise_decrementer= noise_counter = counter;
  } else {
    noise_probability   = probability;
  }
  if( duration_usec >= 0.0 ) {
    noise_duration_usec = duration_usec;
  }

  __selfish_detour_init();
}


static void stop_itimer()
{
  struct itimerval a;
  struct timeval t1, t2;
  int x;

  /* Set the timer to zero, turning everything off */
  t1.tv_sec=0; t1.tv_usec = 0;  a.it_interval=t1;
  t2.tv_sec=0; t2.tv_usec = 0;  a.it_value=t2;
  x = setitimer(ITIMER_REAL, &a, 0);
}

void __selfish_detour_init()
{
  struct itimerval a;
  struct timeval t1, t2;
  int myrank;
  long long ticks_tmp;
  unsigned signal_overhead_ticks=0;
  static int bginfo_flag = 0;


  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

  signal_overhead_ticks=selfish_detour_get_signal_overhead_ticks();
  if( signal_overhead_ticks==0 ) {
    local_printf("Unlikely!! signal overhead was zero\n");
  } else if( signal_overhead_ticks/TICKS_PER_US > 100 ) {
    local_printf("Unlikely!! signal overhead was too big!! It was %u usec\n",
		 signal_overhead_ticks/TICKS_PER_US );
  } else {
    local_printf("# Noise injection test .... passed\n");
    local_printf("# SIGNAL_OVERHEAD_TICKS=%u\n",signal_overhead_ticks);
    local_printf("\n");
  }

  if( noise_interval_usec < 0.01 ) { /* 0.01 is a tentaive value for */
    local_printf("Warning: noise interval is too small. Injection is cancelled.\n");
    stop_itimer();
    MPI_Barrier(MPI_COMM_WORLD);
    return ;
  }
  if( noise_interval_usec >= 1000*1000 ) {
    printf("Error: interval should be less than 1sec! It was %f\n",
         noise_interval_usec);
    exit(1);
  }

  if( noise_probability > 1.0 ) {
    printf("Error: probability should be less equal than 1.0!\n");
    exit(1);
  }


  /* XXX: this is BG/L specific. anyway this code is BG/L specific */
  if(  !( noise_duration_usec <= (double)((unsigned)-1)/(double)TICKS_PER_US) ) {
    printf("Error: detour duration should be less than %f usec!\n",
		 (double)((unsigned)-1)/(double)TICKS_PER_US
		 );
    exit(1);
  }


  if( noise_probability == 0.0 && noise_counter==0 ) {
    local_printf("Warning: since noise_probablity or noise_counter is zero, no noise will be injected.\n");
    stop_itimer();
    MPI_Barrier(MPI_COMM_WORLD);
    return ;
  }

  if( selfish_detour_verbose>0 ) {
    if( bginfo_flag < 1 ) {
      void bginfo_init();
      bginfo_init();   /* must be called from all procs */
    }
  }

  if( (selfish_detour_verbose>0) && myrank == 0 ) {

    local_printf("# NOISE_INJECTION_INTERVAL_USEC=%.f\n",
	   noise_interval_usec);
    if( noise_counter>0 ) {
      local_printf("# NOISE_INJECTION_COUNTER=%d\n",
	     noise_counter);
    } else {
      local_printf("# NOISE_INJECTION_PROBABILITY=%.3f\n",
	     noise_probability);
    }
    local_printf("# NOISE_INJECTION_DURATION_USEC=%.f\n",
		 noise_duration_usec);

    local_printf("# NOISE_INJECTION_sync=%d\n",
	   detour_sync);


    if( bginfo_flag < 1 ) {
      void bginfo_print_info(); 
      bginfo_print_info(); 
      bginfo_flag++;
    }
  }

  /* set the period for wakeup */
  /* XXX: current interval should be less than 1 sec */
  t1.tv_sec=0; t1.tv_usec = noise_interval_usec;
  t2.tv_sec=0; t2.tv_usec = noise_interval_usec;

  a.it_interval=t1;
  a.it_value=t2;

  /*
    We use the ``Mersenne Twister'' random number generator MT19937
    You can seed with any uint32, but the best are odds in 0..(2^32 - 1)

    To make sure that each node starts their random sequence at a 
    different place, we use the rank in the seed 
  */

  if( !detour_sync ) {
    
    seedMT(4357 + (myrank + myrank));  /* this should be an odd number */
  }

  MPI_Barrier(MPI_COMM_WORLD);
  
  /* 
     random delay(up to 1sec) before benchmark
     this affects the start offset of the interval timer
  */
  if( !detour_sync ) {
    unsigned long long delay_ticks = randomMT() % (TICKS_PER_US*1000*1000);
    unsigned long long s,b;
    double x = 0.0;

    s = rdtsc();
    b = s;
    while ( (b-s) < delay_ticks) {
      x++;
      b = rdtsc();
    }
  }

  /* Note: sigaction is not implemented on the BG/L CNK */
#ifndef BLRTS_DECREMENTER_HIJACK
  signal( SIGALRM, selfish_detour );
#else
  hijack_decrementer_handler( (void (*)() ) selfish_detour );
#endif

  if( (unsigned)(noise_duration_usec*(double)TICKS_PER_US) < signal_overhead_ticks ) {
    ticks_tmp = 0;
  } else {
    ticks_tmp = noise_duration_usec*TICKS_PER_US - signal_overhead_ticks;
  }

  if( ticks_tmp < 0 || ticks_tmp > (long long)((unsigned)-1) ) {
    printf("Error: busy loop ticks is wrong :  %s(%d)\n",
		 __FILE__, __LINE__ );
    exit(1);
  }
  detour_loop_ticks = (unsigned long)ticks_tmp;

  if( noise_counter > 0 ) {
    noise_probability_uint32 = 0;
  } else {
    noise_probability_uint32 = 
      (unsigned long)( noise_probability* (double)((unsigned)-1) );
  }

  noise_decrementer= noise_counter;

  /* check internal variables */
  local_printf("# [selfish_detour_proc internal value]\n");
  local_printf("# detour_loop_ticks=%lu\n", detour_loop_ticks);
  local_printf("# noise_probability_uint32=%lu\n", noise_probability_uint32);
  local_printf("# noise_counter=%d\n", noise_counter);

  selfish_detour_last_tick = 0;
  selfish_too_short_interval_count = 0;

  reset_detour_stats();

  local_printf("# before setitimer()\n");
  setitimer(ITIMER_REAL, &a, 0);
  local_printf("# after setitimer()\n");

  MPI_Barrier(MPI_COMM_WORLD);
}


/* Finalizing routine 
   It just stops the interval timer.
*/
void selfish_detour_finalize() 
{
#ifdef COLLECTIVE_SUMMARY
  double min, max, sum;
  double tmp;
  int   size ;
  MPI_Comm_size(MPI_COMM_WORLD, &size);
#endif

  stop_itimer();


  local_printf("# [detour signal stats]\n");

  local_printf("# global_flop          = %.0f\n",
	       global_flop);
  
  local_printf("# detour_signal_cnt    = %llu\n",
	       detour_signal_cnt);

  local_printf("# detour_elapsed_ticks = %llu\n",
	       detour_elapsed_ticks);

#ifdef COLLECTIVE_SUMMARY
  tmp = global_flop;
  MPI_Allreduce( &tmp, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  MPI_Allreduce( &tmp, &min, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
  MPI_Allreduce( &tmp, &max, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

  local_printf("# global_flop          = %.0f(%.0f,%.0f)\n",  
	       sum/(double)size, min,max);


  tmp = (double)detour_signal_cnt;
  MPI_Allreduce( &tmp, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  MPI_Allreduce( &tmp, &min, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
  MPI_Allreduce( &tmp, &max, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

  local_printf("# detour_signal_cnt    = %.0f(%.0f,%.0f)\n", 
	       sum/(double)size, min, max );

  tmp = (double)detour_elapsed_ticks;
  MPI_Allreduce( &tmp, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  MPI_Allreduce( &tmp, &min, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
  MPI_Allreduce( &tmp, &max, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

  local_printf("# detour_elapsed_ticks = %.0f(%.0f,%.0f)\n", 
	       sum/(double)size, min, max );
#endif


  local_printf("\n");

  stupidfence[0] = stupidfence[0];  /* to avoid warning msg */
}
