//***************************************************************************
// SELFISHDETOUR:selfishdetour-info
//     This file is part of Selfish Detour: OS detour(noise) benchmark
// SELFISHDETOUR:selfishdetour-info
//
// SELFISHDETOUR:selfishdetour-fillin
//     $Id: StatsCalc.cpp,v 1.30 2007/01/03 21:16:49 kazutomo Exp $
//     SelfishDetour_Version: 1.0.7
//     SelfishDetour_Heredity: FOSS_ORIG
//     SelfishDetour_License: GPL
// SELFISHDETOUR:selfishdetour-fillin
//
// SELFISHDETOUR:selfishdetour-gpl
//      Copyright: Argonne National Laboratory, Department of Energy,
//                 and University of Chicago.  2004, 2005, 2006, 2007
//      SelfishDetour License: GPL
// 
//      This software is free.  See the file license-files/license.GPL
//      for complete details on your rights to copy, modify, and use this
//      software.
// SELFISHDETOUR:selfishdetour-gpl
//***************************************************************************
/**
 *  
 * StatsCalc contains top-level routine to calculate for detour data.
 * using DetourData call, StatsBasic class
 */

#include <iostream>
#include <fstream>
#include <iomanip>  // for std::setw()
#include <string>
#include <vector>
#include <iterator>
#include <limits>   // numeric_limits
#include <sstream>  // stringstream

#include "StatsCalc.h"
#include "DetourRawData.h"
#include "StatsBasic.h"
#include "UniRand.h"


/**
 * convert number to string
 */
template <typename T>
static std::string _numtoa(T num)
{
  std::stringstream conv;
  conv << num;
  return conv.str();
}


/**
 * make a copy of src and store it to dst
 *
 */
static void copy_sort( const StatsBasic::Vec_t& src, 
		       StatsBasic::Vec_t& dst )
{
  dst.resize( src.size() );
  std::copy( src.begin(), src.end(), dst.begin() );
  std::sort( dst.begin(), dst.end() );
}

/**
 *
 */
static void convertToRank( const StatsBasic::Vec_t& src_v,
			   StatsBasic::Vec_t& dst_v,
			   int npairs = 0 )
{
  dst_v.clear();

  if( npairs>1 ) {
    StatsBasic::Vec_t   tmp_v;
    copy_sort( src_v, tmp_v );
    
    StatsBasic::Vec_t::const_iterator it;

    int localcnt = 0;
    double localsum=0.0;

    for( it=tmp_v.begin(); it!=tmp_v.end(); ++it ) {
      localsum += *it;
      localcnt ++;
      if( localcnt  == npairs ) {

	double av = ( localsum / (double)npairs);
	for( int j=0;j<localcnt; j++ )  dst_v.push_back( av );

	localcnt = 0;
	localsum = 0.0;
      }
    }
    if( localcnt != 0 ) {
      double av = ( localsum / (double)npairs);
      for( int j=0;j<localcnt; j++ )  dst_v.push_back( av );
    }
  } else {
    copy_sort( src_v, dst_v );
  }
}



class LatencySimulation
{
public:
  LatencySimulation( DetourData&  data )
  {
    data.assignDetourDurationInSec( detour_v_m );

    totaltime_m = data.totalTimeLength();

    StatsBasic   detour_s( detour_v_m );
    threshold_m = ( (double)detour_s.total()/totaltime_m );
  }

  ~LatencySimulation( ) {}

  void simulate( int nprocs, 
		 int npairs=0, 
		 int trycount=30 );

  double average()    { return maxminstat_m.average(); }
  double sdeviation() { return maxminstat_m.sdeviation(); }

  double threshold()  { return threshold_m; }

private:
  StatsBasic::Vec_t   detour_v_m;
  double              totaltime_m;

  StatsBasic::Vec_t   sorted_detour_v_m;
  StatsBasic::Vec_t   maxmin_m;

  double              threshold_m;

  StatsBasic          maxminstat_m;

  UniRand rnd_m;   // this is not good because every class construction,
		 // rnd is initialized
};

/**
 *
 *
 */
void LatencySimulation::simulate( int nprocs, 
				  int npairs,
				  int trycount )
{
  /**
   * need to reconstruct every time when simulate() is called
   */
  convertToRank( detour_v_m , sorted_detour_v_m, npairs );

  /*
  StatsBasic aa( sorted_detour_v_m );
  std::cout << "----------" << std::endl;
  std::cout << "nprocs=" << nprocs << 
    " / npairs=" << npairs << std::endl;
  std::cout << "max=" << aa.max() << std::endl;
  double  maxmax = aa.max();
  */

  StatsBasic::Vec_t  x;

  maxmin_m.clear();

  /**
   * try simualtion specified times
   */
  for( int tr=0; tr< trycount; tr ++ ) {
    x.clear();

    for( int cpu=0; cpu<nprocs; cpu++ ) {
      double s;
      double r = rnd_m.rand();

#if 0
      if( r < threshold_m ) {
	int idx;
	/**
	 * r = [0, threshold)
	 * thus, r/threshold = [0, 1.0)
	 * finally, idx = [0,  sorted_detour_v.size() ) 
	 */
	idx = (int)( sorted_detour_v_m.size()*r/threshold_m  );
	s = sorted_detour_v_m[idx];
      } else {
	s = 0.0;
      }
#endif
      int idx = (int)( sorted_detour_v_m.size()*r );
      s = sorted_detour_v_m[idx];
      
      x.push_back( s );
    }
    StatsBasic xstat(x);

    maxmin_m.push_back(  (xstat.max()-xstat.min())  );
  }

  maxminstat_m.calc(maxmin_m);

  /*
  std::cout << "simulated max=" << maxminstat_m.max() << std::endl;

  if( maxminstat_m.max() > maxmax ) {
    std::cout << "something wrong" << std::endl;
  }
  */
}




static class SimulationSetting
{
  public:

  SimulationSetting(void)

  {
    nodes_m.push_back(     64);  // 2^6
    nodes_m.push_back(    256);  // 2^8
    nodes_m.push_back(   1024);  // 2^10
    nodes_m.push_back(   4096);  // 2^12
    nodes_m.push_back(  16384);  // 2^14
    nodes_m.push_back(  65536);  // 2^16
    nodes_m.push_back( 262144);  // 2^18


#if OLDSETTING
    nodes_m.push_back(      2);  // 2^1
    nodes_m.push_back(     16);  // 2^4
    nodes_m.push_back(    256);  // 2^8
    nodes_m.push_back(   1024);  // 2^10
    nodes_m.push_back(   4096);  // 2^12
    nodes_m.push_back(  16384);  // 2^14
    nodes_m.push_back(  65536);  // 2^16
    nodes_m.push_back(1048576);  // 2^20
    nodes_m.push_back(16777216);  // 2^24
#endif
    /**
     * NOTE: 1=2^0 doesn't work due to calculation routine of
     * variation requires N>1 data.
     */

    // Q or 1/SYNCFREQ
    // lengths_m.push_back( pow(10.0,-3.5) );

#if 0
    lengths_m.push_back( 0.001 );
    lengths_m.push_back( 0.01  );
    lengths_m.push_back( 0.1   );
    lengths_m.push_back( 1 );
#endif


    lengths_m.push_back( pow(10.0,-2.5) );
    lengths_m.push_back( pow(10.0,-1.5) );
    lengths_m.push_back( pow(10.0,-0.5) );
    lengths_m.push_back( pow(10.0, 0.5) );


  }
  ~SimulationSetting() {}

  int nodes(int i) { return nodes_m[i]; }
  double lengths(int i) { return lengths_m[i]; }

  int n_nodes() { return nodes_m.size(); }
  int n_lengths() { return lengths_m.size(); }

  std::vector<int> nodes_m;
  std::vector<double> lengths_m;

} setting;



/**
 * do simulation for parallel machine based on detour recoring data
 *
 */
static void parallel_simulation( DetourData& data,
				 AnalyzeDetourOptions& opts )
{
  StatsBasic::Vec_t   detour_v;
  StatsBasic  detour_s;
  StatsBasic  run_s;

  std::string basename = opts.outputbn();


  std::string   nrz_output_fn =  basename  + "-3d.dat";
  std::ofstream f_nrz_output( nrz_output_fn.c_str(), std::ios::out);

  std::string nrz_output_multi_fn = basename + "-2d.dat";
  std::ofstream f_nrz_output_multi( nrz_output_multi_fn.c_str(), std::ios::out);

  f_nrz_output << "# [S.FREQ]   [NNODES] [AV/Q]          [SD/Q]" << std::endl;

  f_nrz_output_multi << "# [S.FREQ]   [NNODES-1] [AV-1/Q] [SD-1/Q] [NNODES-2] [AV-2/Q] [SD-2/Q] ... " << std::endl;


  for( int idx_l=0;  idx_l < setting.n_lengths(); idx_l++ ) {
    DetourData  targetdata( data );

    if( targetdata.resample( setting.lengths(idx_l) ) == false ) {
      break;
    }


    targetdata.assignDetourDurationInSec( detour_v );

    detour_s.calc( detour_v );

    
    double qval = setting.lengths(idx_l);
    double syncfreq = 1.0/qval;
    
    std::string report_title = 
      "\n------ [resampled with Q=" 
      + _numtoa(qval) + "] ----------------------";

    detour_s.report( report_title.c_str() );

    LatencySimulation sim( targetdata );


    f_nrz_output_multi.setf(std::ios::left);
    f_nrz_output_multi << std::setw(10) << syncfreq << " ";

    for( int nnodes=0; nnodes < setting.n_nodes() ; nnodes++ ) {

      /** do parallel calculation with node setting.
	  it will take some time */
      sim.simulate( setting.nodes( nnodes ) );

      /** 
	  output data to file 
      */
      f_nrz_output.setf(std::ios::left);
      f_nrz_output << std::setw(10) << syncfreq << " ";
      f_nrz_output.unsetf(std::ios::left);
      f_nrz_output << std::setw(8) << setting.nodes( nnodes ) << "  " ;
      f_nrz_output.setf(std::ios::left);
      f_nrz_output << std::setw(12)<< (sim.average()/qval) << " " 
		   << std::setw(12)<< (sim.sdeviation()/qval) << std::endl;

      f_nrz_output_multi << setting.nodes( nnodes ) << "  " ;
      f_nrz_output_multi << (sim.average()/qval) << " ";
      f_nrz_output_multi << (sim.sdeviation()/qval) << "   ";
    }

    f_nrz_output_multi << std::endl;
  }

  f_nrz_output.close();
  f_nrz_output_multi.close();
}


/**
 *
 */
static void writeStatsResult( DetourData& data,
			      AnalyzeDetourOptions& opts )
{
  /**
   * stats results for original input data
   */

  std::string statsres_fn = opts.outputbn()+"-stats.dat";
  std::ofstream statsres( statsres_fn.c_str() , std::ios::out);
  
  statsres << "# Basic Stats Result for data" << std::endl;
  statsres << "# Unit for time is second" << std::endl;
  statsres << std::endl;

  StatsBasic::Vec_t   detour_v;
  data.assignDetourDurationInSec( detour_v );

  statsres << "TOTAL_RECORDING_TIME=" <<
    data.totalTimeLength() << std::endl;


  data.report(statsres);

  StatsBasic  detour_s;
  detour_s.calc( detour_v );
  detour_s.report( "DETOUR_", statsres );

  /**
   * calculate min and max interval between detours
   * it can be calculated from run segs
   * min is relatively important
   */
  
  StatsBasic::Vec_t   run_v;
  StatsBasic  run_s;

  data.assignRunDurationInSec( run_v );
  run_s.calc( run_v );
  run_s.report( "RUN_" , statsres );

  statsres << "DETOUR_RUN_RATIO=" <<
    (detour_s.total() / data.totalTimeLength()) << std::endl;

//    detour_s.total() / run_s.total() << std::endl;

  statsres.close();


  

}


static void writeConstructedData( DetourData& data,
				  AnalyzeDetourOptions& opts )
{
  /**
   * write timeseries data and detour sorted data 
   * unit is second
   */
  std::string sorted_detour_fn = opts.outputbn() + "-sorted.dat";
  data.writeDetourSort( sorted_detour_fn.c_str() ); 

  std::string _fn = opts.outputbn() + "-sorted-nx.dat";
  data.writeDetourSortPercent( _fn.c_str(), false ); 

  _fn = opts.outputbn() + "-sorted-nxy.dat";
  data.writeDetourSortPercent( _fn.c_str(), true ); 





  std::string timeseries_detour_fn = opts.outputbn() + "-timeseries.dat";
  data.writeDetourTimeSeries( timeseries_detour_fn.c_str() );

  

}


/**
 *
 */
int statscalc( AnalyzeDetourOptions& opts )
{
  DetourData  data;

  /**
   * Reads data from a file or stdin 
   */
  data.parse( opts.datafn() );

  writeStatsResult( data, opts );

  writeConstructedData( data, opts );

  if( opts.enableSimulation() ) {
    parallel_simulation(data, opts );
  }

  return 0;
}
